import java.util.HashMap;
import java.util.Map;

/**
 * A decision tree class meant to work with HaarFeatures.
 * @author Lekan Wang (lekan@lekanwang.com)
 *
 */
public class DecisionTree {
	/**
	 * Represents a single decision node in the Decision Tree.
	 * @author Lekan Wang (lekan@lekanwang.com)
	 *
	 */
	public class DecisionNode {
		// Optional Name (could be useful for debugging)
		String name;
		
		// The feature to split on
		HaarFeature feature;
		
		// The value to split on
		int split;
		
		// What should the upper half of the split predict? (true=pos; false=neg)
		boolean upperSplitPrediction;
		
		// Child decision nodes
		DecisionNode falseChild, trueChild;
		
		public DecisionNode(String name, HaarFeature feature, int split, boolean upperSplitPrediction) {
			this(feature, split, upperSplitPrediction);
			this.name = name;
		}
		
		public DecisionNode(HaarFeature feature, int split, boolean upperSplitPrediction) {
			this.name = ("" + autoID++);
			this.feature = feature;
			this.split = split;
			this.upperSplitPrediction = upperSplitPrediction;
			this.falseChild = this.trueChild = null;
		}
	}

	private static int autoID = 0;
	private DecisionNode root;
	
	Map<String, DecisionNode> nodeMap;
	
	
	public DecisionTree() {
		root = null;
		nodeMap = new HashMap<String, DecisionNode>();
	}
	
	/**
	 * Same as the other overloaded addNode method, but allows giving the new node a
	 * custom name, which can be used for debugging and
	 * other purposes.
	 * @param parent
	 * @param parentSplit
	 * @param feature
	 * @param split
	 * @param upperSplitPrediction
	 * @param name
	 */
	public void addNode(DecisionNode parent, boolean parentSplit, HaarFeature feature,
			int split, boolean upperSplitPrediction, String name) {
		DecisionNode newNode = name==null ? 
				new DecisionNode(feature, split, upperSplitPrediction) :
				new DecisionNode(name, feature, split, upperSplitPrediction);
		
		if (parent == null)
			root = newNode;
			
		if (parentSplit) {
			parent.trueChild = newNode;
		} else {
			parent.falseChild = newNode;
		}
		
		nodeMap.put(newNode.name, newNode);
	}
	
	/**
	 * Adds a new decision node to the tree, with the given parameters. If the parent is null,
	 * will create a new root node. Will not do any checking to prevent orphaning of nodes, nor
	 * replacing existing nodes.
	 * 
	 * In the case of a new root node, the parentSplit is obviously not used, so just put in a
	 * placeholder true or false.
	 * 
	 * @param parent The parent decisionNode
	 * @parem parentSplit Whether this is the child for when the parent predicts positive (true), or for when the parent predicts negative (false)
	 * @param feature The Haar-like feature associated with this decision node
	 * @param split The value to split on
	 * @param upperSplitPrediction What the upper half of the split predict
	 */
	public void addNode(DecisionNode parent, boolean parentSplit, HaarFeature feature,
			int split, boolean upperSplitPrediction) {
		addNode(parent, parentSplit, feature, split, upperSplitPrediction, null);
	}
	
	/**
	 * Returns the root of this DecisionTree
	 * @return
	 */
	public DecisionNode getDecisionTree() {
		return this.root;
	}
	
	/**
	 * Returns the DecisionNode associated with a given name.
	 * 
	 * @param name
	 * @return
	 */
	public DecisionNode getNode(String name) {
		return nodeMap.get(name); 
	}
}
